"""Web app forms."""

from flask import current_app as app
from flask_table import Col, Table
from flask_wtf import FlaskForm
from steam import game_servers as gs
from wtforms import (
    BooleanField,
    IntegerField,
    RadioField,
    SelectMultipleField,
    StringField,
    SubmitField,
    validators,
    widgets,
)

# -------------------------------------------------------------------------------


class ServerTable(Table):
    """ServerTable."""

    classes = ["table", "table-sm", "table-condensed", "table-bordered"]

    region = Col("Region")
    app_id = Col("Appid")
    server_type = Col("ST")
    vac = Col("Vac")
    visibility = Col("Vis")
    n_imposters = Col("Imp", column_html_attrs={"class": "text-right"})
    server_host = Col("Host")
    server_port = Col("Port")
    ping = Col("Ping", column_html_attrs={"class": "text-right"})
    players = Col("Players", column_html_attrs={"class": "text-right"})
    max_players = Col("Max", column_html_attrs={"class": "text-right"})
    bots = Col("Bots", column_html_attrs={"class": "text-right"})
    map_name = Col("Map")
    server_name = Col("Name")
    # keywords        = Col('Keywords')

    # -------------------------------------------------------------------------------

    def get_tr_attrs(self, item):
        """Return table row attributes for `item`."""

        classes = "server"
        if (attrs := item.get("tr_attrs")) is not None:
            classes += " " + attrs

        return {
            "class": classes,
            "addr": item["addr"],
        }

    # -------------------------------------------------------------------------------

    def sort_url(self, col_id, reverse=False):
        """W0223.

        Method 'sort_url' is abstract in class 'Table' but is not
        overridden (abstract-method).
        """


# -------------------------------------------------------------------------------
# other available properties
#   '_ping',        # 108.20508003234863
#   '_type',        # 'source'
#   'edf',          # 177
#   'environment',  # 'l'
#   'folder',       # 'tf'
#   'game',         # '\x01  ►  Gun Game  ◄'
#   'game_id',      # 440
#   'keywords',     # 'alltalk,blackwonder,cp,game,gg,gun,gungame,ranks'
#   'port',         # 27075
#   'protocol',     # 17
#   'steam_id',     # 85568392921190943
#   'version',      # '6300758'

# -------------------------------------------------------------------------------


class MultiCheckboxField(SelectMultipleField):
    """MultiCheckboxField."""

    widget = widgets.ListWidget(prefix_label=False)
    option_widget = widgets.CheckboxInput()


# -------------------------------------------------------------------------------


class RadioBool(RadioField):
    """RadioBool."""

    def __init__(self, label="", validators_=None, **kwargs):
        """Initialize RadioBool."""

        kwargs["choices"] = [("None", "Unspecified"), ("True", "True"), ("False", "False")]
        if not kwargs.get("default"):
            kwargs["default"] = "None"
        kwargs["render_kw"] = {"class": "btn-check"}
        super().__init__(label, validators_, **kwargs)


# -------------------------------------------------------------------------------


class SearchForm(FlaskForm):
    """SearchForm."""

    # search engine
    max_threads = IntegerField(
        "Max Threads",
        [validators.NumberRange(1, 100, "Please enter a number from 1 to 100")],
        default=app.config["args"].max_threads,
        render_kw={"size": 4},
    )
    debug = BooleanField("Debug", default=app.config["args"].debug)

    # stage1 filters
    max_servers = IntegerField(
        "Max Servers", default=app.config["args"].max_servers, render_kw={"size": 4}
    )
    regions = MultiCheckboxField(
        "Regions",
        choices=[(str(x.value), x.name) for x in gs.MSRegion],
        default=[
            str(x.value)
            for x in (
                gs.MSRegion.US_East,
                gs.MSRegion.US_West,
                gs.MSRegion.South_America,
                gs.MSRegion.Europe,
                # gs.MSRegion.Asia.value,
                # gs.MSRegion.Australia.value,
                # gs.MSRegion.Middle_East.value,
                # gs.MSRegion.Africa.value,
                # gs.MSRegion.World.value,
            )
        ],
    )

    appid = IntegerField("AppID", [validators.optional()], default=440, render_kw={"size": 4})
    dedicated = RadioBool("Dedicated")
    secure = RadioBool("Secure")
    password = RadioBool("Password")
    empty = RadioBool("Not Empty", default="True")
    full = RadioBool("Not Full", default="True")
    noplayers = RadioBool("No Players")
    map_name = StringField(
        "Map name", default=app.config["args"].map_name, render_kw={"size": 30}
    )
    map_prefix = StringField(
        "Map prefix", default=app.config["args"].map_prefix, render_kw={"size": 30}
    )

    # stage2 filters go here

    min_players = IntegerField("Min Players", [validators.optional()], render_kw={"size": 4})
    max_bots = IntegerField("Max Bots", [validators.optional()], render_kw={"size": 4})
    max_ping = IntegerField("Max Ping", [validators.optional()], render_kw={"size": 4})

    #
    submit = SubmitField("Search", render_kw={"class": "btn btn-primary btn-lg btn-block"})


# -------------------------------------------------------------------------------



"""Valve Main Server."""

# -------------------------------------------------------------------------------

import queue
import threading

from loguru import logger
from steam import game_servers as gs

import qvalve.gameserver

# -------------------------------------------------------------------------------


class MainServer:
    """Valve Main Server.

    Interface to 1) search Valve's Main Server for list of Remote Game
    Servers, and 2) multi-thread querying of Remote Game Servers for
    current map, number of players, etc.

    See https://github.com/ValvePython/steam, which is an interface to
    https://developer.valvesoftware.com/wiki/Master_Server_Query_Protocol.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, max_threads=10, debug=False, rules=False):
        """Initialize MainServer."""

        self._max_threads = int(max_threads)
        self._debug = bool(debug)
        self._rules = bool(rules)
        self._workq = None

    # -------------------------------------------------------------------------------
    # query_master(
    #   filter_text='\\nappid\\500',
    #   max_servers=20,
    #   region=<MSRegion.World: 255>,
    #   master=('hl2master.steampowered.com', 27011),
    #   timeout=2)

    def search(self, regions, **kwargs):
        """Query valve's main server.

        For each region in the list of `regions`, query the main server for
        a list of remote game servers that meet criteria in `filters`,
        which may be a string or a dict. Return `list(dict)`.
        """

        if self._workq is None:
            threading.current_thread().name = "main"
            self._workq = queue.Queue()
            for _ in range(self._max_threads):
                threading.Thread(target=self._a2s_worker, daemon=True).start()

        # `query_master` wants `filter_text` as `type(str)`;
        # also accept `filters` as `type(dict)`.

        if (filters := kwargs.get("filters")) is not None and isinstance(filters, dict):
            delim = "\\"
            filters = delim.join([f"{k}{delim}{v}" for k, v in filters.items()])
            kwargs["filter_text"] = filters
            del kwargs["filters"]

        # build and return this list.
        servers = []

        for region in regions:

            if not isinstance(region, gs.MSRegion):
                region = gs.MSRegion(int(region))
            kwargs["region"] = region
            logger.debug(kwargs)

            # Search Valve's Main Server for list of Remote Game Servers
            # with matching criteria.
            #
            # Create a `GameServer` for each item returned, but don't query
            # them from here (main thread); send them through the workq to
            # workers to perform.

            try:
                for addr in gs.query_master(**kwargs):
                    gameserver = qvalve.gameserver.GameServer(addr, region.value)
                    servers.append(gameserver)
                    self._workq.put(gameserver)

            except RuntimeError as err:
                logger.error(f"{err} query_master({kwargs}")

        # wait for all threads to drain the queue
        self._workq.join()

        # return servers we were able to ping
        return [x for x in servers if x.ping is not None]

    # -------------------------------------------------------------------------------

    def _a2s_worker(self):

        while True:
            try:
                gameserver = self._workq.get()
            except queue.Empty:
                return

            try:
                gameserver.query()
            finally:
                self._workq.task_done()


# -------------------------------------------------------------------------------



"""Reports.

Functions to print text reports to stdout.
"""

# -------------------------------------------------------------------------------

from collections import defaultdict

from loguru import logger

import qvalve.gameserver
import qvalve.mainserver

# -------------------------------------------------------------------------------


def search_mainserver(args):
    """Search Valve's Main server for Game servers."""

    mainserver = qvalve.mainserver.MainServer(
        max_threads=args.max_threads,
        debug=args.debug,
        rules=args.show_tags,
    )

    filters = _get_filters_stage1(args)

    servers = mainserver.search(
        regions=args.regions, filters=filters, max_servers=args.max_servers
    )
    logger.success(f"mainserver.search returned {len(servers)} servers")

    servers = _filter_stage2(args, servers)
    logger.success(f"_filter_stage2 returned {len(servers)} servers")

    # key = lambda x: (-x['players'], x['mapname'])  # noqa: E731
    # key = lambda x: (x['mapname'], -x['players'])  # noqa: E731

    _print_gameservers(
        args,
        sorted(servers, key=lambda x: (x.map_name, x.ping, x.addr)),
    )

    if args.report_keywords:
        _print_keywords_report(servers)


# -------------------------------------------------------------------------------


def query_gameservers(args):
    """Query list of Game server addresses."""

    for addr in args.addrs:
        server = qvalve.gameserver.GameServer(addr)
        server.query()
        _print_gameservers(args, [server])


# -------------------------------------------------------------------------------


def _get_filters_stage1(args):
    """Sent to valve in query to get list of remote game servers."""

    filters = {}
    # region is added elsewhere
    if args.appid is not None:
        filters["appid"] = args.appid
    # if args.dedicated is not None:
    #    filters['dedicated'] = args.dedicated
    # if args.secure is not None:
    #    filters['secure'] = args.secure
    # if args.password is not None:
    #    filters['password'] = args.password
    if args.empty is not None:
        filters["empty"] = args.empty
    if args.full is not None:
        filters["full"] = args.full
    if args.noplayers is not None:
        filters["noplayers"] = args.noplayers
    if args.map_name is not None:
        filters["map_name"] = args.map_name

    return filters


# -------------------------------------------------------------------------------


def _filter_stage2(args, servers):
    """Applied after querying valve."""

    if args.map_prefix is not None:
        _ = [x for x in servers if x.map_name.startswith(args.map_prefix)]
        count = len(_)
        logger.info(f"removed {len(servers) - count} servers leaving {count}; map_prefix")
        servers = _

    if args.min_players is not None:  # int
        _ = [x for x in servers if x.players >= args.min_players]
        count = len(_)
        logger.info(f"removed {len(servers) - count} servers leaving {count}; min_players")
        servers = _

    if args.no_max_players:  # bool
        _ = [x for x in servers if x.players < x.max_players]
        count = len(_)
        logger.info(f"removed {len(servers) - count} servers leaving {count}; no_max_players")
        servers = _

    if args.no_mm_strict_1:  # bool
        _ = [x for x in servers if x.visibility != 1]
        count = len(_)
        logger.info(f"removed {len(servers) - count} servers leaving {count}; no_mm_strict_1")
        servers = _

    if args.max_ping is not None:  # int
        _ = [x for x in servers if x.ping <= args.max_ping]
        count = len(_)
        logger.info(f"removed {len(servers) - count} servers leaving {count}; max_ping")
        servers = _

    return servers


# -------------------------------------------------------------------------------


def _print_gameservers(args, servers):

    lastmap = None

    for server in [x for x in servers if x.map_name]:

        if lastmap and lastmap != server.map_name:
            print("-" * 150)
        lastmap = server.map_name

        print(
            " ".join(
                [
                    f"r={server.region:1}",
                    f"app_id={server.app_id:3}",
                    f"typ={server.server_type:1}",
                    f"vac={server.vac:1}",
                    f"vis={server.visibility:1}",
                    f"imp={server.n_imposters:2}",
                    f"a={server.addr:21}",
                    f"ping={server.ping:3}",
                    f"p={server.players:2}",
                    f"m={server.max_players:2}",
                    f"b={server.bots:2}",
                    f"{server.map_name:35}",
                    f"{server.server_name!r}",
                ]
            )
        )

        if args.show_keywords:
            print(f"keywords={server.keywords!r}")

        if args.show_tags:
            print(f".sv_tags={server.sv_tags!r}")

        if args.show_tags and args.show_keywords:
            if not server.keywords or not server.sv_tags:
                logger.warning("missing")
            elif server.sv_tags != server.keywords:
                logger.error("mismatch")
            else:
                logger.success("match")

        if args.show_players:
            for name in server.playernames:
                print(" " * 4 + repr(name))

        for player in server.known_hackers:
            print(player)


# -------------------------------------------------------------------------------


def _print_keywords_report(servers):

    nplayers = defaultdict(int)
    nservers = defaultdict(int)

    for server in [x for x in servers if x.keywords]:
        for k in server.keywords:
            nservers[k] += 1
            nplayers[k] += server.players

    # print('Keywords by number of players:')
    # for key, count in sorted(nplayers.items(), key=lambda _: _[1], reverse=True):
    #    print(f'{count:3} {key}')

    # print('Keywords by number of servers:')
    # for key, count in sorted(nservers.items(), key=lambda _: _[1], reverse=True):
    #    print(f'{count:3} {key}')

    count = 30
    print(f"Top {count} nplayers, nservers")
    _a = sorted(nplayers.items(), key=lambda _: _[1], reverse=True)[:count]
    _b = sorted(nservers.items(), key=lambda _: _[1], reverse=True)[:count]
    for _ in zip(_a, _b):
        print(f"{str(_[0]):30} {str(_[1]):30}")


# -------------------------------------------------------------------------------



"""Command line interface."""

import logging
import sys
from pathlib import Path

# import tf2mon.hacker
from libcli import BaseCLI
from loguru import logger
from steam.game_servers import MSRegion

import qvalve.flaskapp
import qvalve.gameserver
import qvalve.reports

__all__ = ["QvalveCLI"]


class QvalveCLI(BaseCLI):
    """Command line interface."""

    config = {
        # name of config file.
        "config-file": Path("~/.qvalve.toml"),
        # toml [section-name].
        "config-name": "qvalve",
        # application
        "gamebots": Path(__file__).parent.joinpath("data/gamebots.csv"),
        "max-threads": 10,
        "max-servers": 100,
    }

    def init_logging(self, verbose: int) -> None:
        """Set loguru logging level based on `--verbose`."""

        if logging.root.handlers:
            return
        super().init_logging(verbose)
        _ = ["INFO", "DEBUG", "TRACE"]
        level = _[min(verbose, len(_) - 1)]
        logger.remove()
        logger.add(sys.stderr, level=level)

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.parser = self.ArgumentParser(
            prog=__package__,
            description=self.dedent("""
    Search `Valve`s Main server for Game servers. Integrated with `tf2mon`s
    hacker-database to identify known cheaters on game servers. Click on a
    server to show/hide its players; `Ctrl-Click` on server (or players) to
    subsequently connect to that server when `F12` is pressed in-game.
                """),
        )

    def set_defaults(self):
        """Set defaults."""

        self.parser.set_defaults(
            max_threads=self.config["max-threads"],
            debug=False,
            show_players=False,
            show_keywords=False,
            show_tags=False,
            report_keywords=False,
            # stage1 filters
            max_servers=self.config["max-servers"],
            regions=[
                MSRegion.US_East.value,
                MSRegion.US_West.value,
                MSRegion.South_America.value,
                MSRegion.Europe.value,
                # MSRegion.Asia.value,
                # MSRegion.Australia.value,
                # MSRegion.Middle_East.value,
                # MSRegion.Africa.value,
                # MSRegion.World.value,
            ],
            appid=440,  # Servers that are running game
            # these have no effect in testing, so disabling.
            # dedicated=None,   # Servers running dedicated
            # secure=None,      # Servers using anti-cheat technology
            # password=None,    # Servers that are not password protected
            empty=None,  # Servers that are not empty
            full=None,  # Servers that are not full
            noplayers=None,  # Servers that are empty
            map_name=None,  # exact match; 'plr_pipeline'
            # stage2 filters
            min_players=None,
            no_max_players=None,
            max_ping=None,
            no_mm_strict_1=None,
            map_prefix=None,  # prefix match; 'plr_'
            # usage 2
            addrs=[],
            # usage 3
            web_server=False,
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        # usage 1
        arg = self.parser.add_argument(
            "--max-threads",
            metavar="NUM",
            type=int,
            help="Run `NUM` threads for game server comms.",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--debug",
            action="store_true",
            help="Pretty-print raw response records",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--show-players",
            action="store_true",
            help="Print `A2S_PLAYER.names`",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--show-keywords",
            action="store_true",
            help="Print `A2S_INFO.keywords`",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--show-tags",
            action="store_true",
            help="Print `A2S_RULES.sv_tags`",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "--report-keywords",
            action="store_true",
            help="Print keywords report",
        )
        self.add_default_to_help(arg)

        # -------------------------------------------------------------------------------

        stage1 = self.parser.add_argument_group(
            "Stage one filters, sent to valve in query to get list of remote game servers"
        )
        arg = stage1.add_argument(
            "--max-servers",
            metavar="NUM",
            type=int,
            help="Get no more than `NUM` servers per region",
        )
        self.add_default_to_help(arg)

        arg = stage1.add_argument(
            "--regions",
            metavar="NUM",
            type=int,
            nargs="+",
            help="Get servers for list of regions",
        )
        self.add_default_to_help(arg)

        arg = stage1.add_argument(
            "--appid",
            metavar="NUM",
            type=int,
            help="Servers that are running game",
        )
        self.add_default_to_help(arg)

        # stage1.add_argument('--dedicated', metavar='NUM', type=int,
        #   help='Servers running dedicated')
        # stage1.add_argument('--secure', metavar='NUM', type=int,
        #   help='Servers using anti-cheat technology')
        # stage1.add_argument('--password', metavar='NUM', type=int,
        #   help='Servers that are not password protected')

        arg = stage1.add_argument(
            "--empty",
            metavar="NUM",
            type=int,
            help="Servers that are not empty",
        )

        arg = stage1.add_argument(
            "--full",
            metavar="NUM",
            type=int,
            help="Servers that are not full",
        )

        arg = stage1.add_argument(
            "--noplayers",
            metavar="NUM",
            type=int,
            help="Servers that are empty",
        )

        arg = stage1.add_argument(
            "--map-name",
            metavar="NAME",
            help="Match map `NAME` (exact)",
        )

        # -------------------------------------------------------------------------------

        stage2 = self.parser.add_argument_group(
            "Stage two filters, applied after querying valve",
        )
        stage1.add_argument(
            "--map-prefix",
            metavar="PREFIX",
            help="Match map names that start with `PREFIX`",
        )

        stage2.add_argument(
            "--min-players",
            metavar="NUM",
            type=int,
            help="where number of players is at least NUM",
        )

        stage2.add_argument(
            "--no-max-players",
            action="store_true",
            help="where number of players is less than its `max_players`",
        )

        stage2.add_argument(
            "--max-ping",
            metavar="NUM",
            type=int,
            help="where ping is NUM or less",
        )

        stage2.add_argument(
            "--no-mm-strict-1",
            action="store_true",
            help="where tf_mm_strict is not 1",
        )

        #
        usage2 = self.parser.add_argument_group("Usage 2")
        usage2.add_argument(
            "addrs",
            metavar="ADDR",
            nargs="*",
            help="Query list of Game server addresses, where ADDR is `IP:PORTNO`",
        )

        #
        usage3 = self.parser.add_argument_group("Usage 3")
        usage3.add_argument(
            "--web-server",
            action="store_true",
            help="Run web server",
        )

    def main(self) -> None:
        """Command line interface entry point (method)."""

        # hackers = tf2mon.hacker.HackerManager()
        # hackers.load_gamebots(self.config["gamebots"])

        class HackerManager:
            """Docstring."""

            # pylint: disable=too-few-public-methods
            def lookup_name(self, name: str):
                """Docstring."""
                logger.info(f"HackerManager.lookup_name: {name!r}")

        hackers = HackerManager()

        qvalve.gameserver.GameServer.configure(self.options, hackers)

        if self.options.web_server:
            # usage 3
            qvalve.flaskapp.run_web_server(self.options)

        elif self.options.addrs:
            # usage 2
            qvalve.reports.query_gameservers(self.options)

        else:
            # usage 1
            qvalve.reports.search_mainserver(self.options)


def main(args: list[str] | None = None) -> None:
    """Command line interface entry point (function)."""
    QvalveCLI(args).main()



import os

import pytest

from qvalve.cli import main


def test_no_args() -> None:
    main(["--regions", "1", "-v"])


slow = pytest.mark.skipif(not os.environ.get("SLOW"), reason="slow")


@slow
def test_map_name() -> None:
    main(["--regions", "1", "-vv", "--map-name", "plr_highertower"])


@slow
def test_map_prefix() -> None:
    main(["--regions", "1", "-vv", "--map-prefix", "plr_"])


@slow
def test_debug() -> None:
    main(["--regions", "1", "-vvv", "--map-name", "pl_upward", "--debug"])


@slow
def test_show_keywords() -> None:
    main(["--regions", "1", "-v", "--map-name", "pl_upward", "--show-keywords"])


@slow
def test_show_players() -> None:
    main(["--regions", "1", "-v", "--map-name", "pl_upward", "--show-players"])


@slow
def test_show_tags() -> None:
    main(["--regions", "1", "-v", "--map-name", "pl_upward", "--show-tags"])


@slow
def test_report_keywords() -> None:
    main(["--regions", "1", "-v", "--report-keywords"])


# def test_web():
#    os.environ['FLASK_APP'] = 'qvalve.web'
#    os.environ['FLASK_ENV'] = 'development'
#    run(
#        args=[
#            'python',
#            '-m',
#            'flask',
#            'run',
#            '--no-debugger',
#        ],
#        check=True,
#    )



"""Remote Game Server."""

# -------------------------------------------------------------------------------

import json
import socket
from pprint import pprint as pp

# from fuzzywuzzy import fuzz
import steam.game_servers
from loguru import logger

# -------------------------------------------------------------------------------


class GameServer:
    """Remote Game Server.

    Interface to query Remote Game Server for current map, number of players, etc.

    See https://github.com/ValvePython/steam, which is an interface to
    https://developer.valvesoftware.com/wiki/Server_queries.
    """

    # pylint: disable=too-many-instance-attributes
    # -------------------------------------------------------------------------------

    _hackerdb = None
    _debug = None
    _rules = None

    @classmethod
    def configure(cls, args, hackerdb):
        """Configure interface."""

        cls._hackerdb = hackerdb
        cls._debug = args.debug
        cls._rules = args.show_tags

    # -------------------------------------------------------------------------------

    def __init__(self, addr, region=None):
        """Initialize new `GameServer`.

        Args:
            addr: yada
            region: yada
        """

        if not self._hackerdb:
            raise RuntimeError("`configure` not called.")

        if isinstance(addr, str):
            host, port = addr.split(":")
            port = int(port)
        elif isinstance(addr, (list, tuple)):  # noqa: SIM106 Handle error-cases first
            host, port = addr
        else:
            raise SyntaxError(f"expecting str or list, not {type(addr)}; addr={addr!r}")

        self.server_host = host
        self.server_port = int(port)
        self.server_addr = (host, port)
        self.addr = f"{host}:{port}"

        #
        self.region = 9 if region is None else region

        # from `steam.game_servers.a2s_info`
        self.app_id = None
        self.server_type = None
        self.vac = None
        self.visibility = None
        self.players = None
        self.max_players = None
        self.bots = None
        self.map_name = None
        self.server_name = None
        self.keywords = []
        # our extensions
        self.ping = None

        # from `steam.game_servers.a2s_player`
        self.a2s_players = []
        # our extensions
        self.playernames = []
        self.known_hackers = []
        self.n_imposters = 0

        # from `steam.game_servers.a2s_rules`
        # <none; sv_tags comes as string; we split and store as list>
        # our extensions
        self.sv_tags = []

    # -------------------------------------------------------------------------------

    def __str__(self):

        string = f"{self.__class__.__name__}({self.addr!r}"
        if self.map_name:
            string += f", {self.map_name!r}"
        if self.server_name:
            string += f", {self.server_name!r}"
        return string + ")"

    # -------------------------------------------------------------------------------

    def to_json(self):
        """Return json repr of self."""

        serializable = {
            key: value for key, value in self.__dict__.items() if key not in ("known_hackers")
        }
        try:
            return json.dumps(serializable)
        except Exception as err:  # pylint: disable=broad-except
            logger.error(err)
            pp(serializable)
            return {}

    #        return json.dumps({
    #            'region': self.region,
    #            'addr': self.addr,
    #            'app_id': self.app_id,
    #            'server_type': self.server_type,
    #            'vac': self.vac,
    #            'visibility': self.visibility,
    #            'n_imposters': self.n_imposters,
    #            'server_host': self.server_host,
    #            'server_port': self.server_port,
    #            'ping': self.ping,
    #            'players': self.players,
    #            'max_players': self.max_players,
    #            'bots': self.bots,
    #            'map_name': self.map_name,
    #            'server_name': self.server_name,
    #            'a2s_players': self.a2s_players,
    #        })

    # -------------------------------------------------------------------------------

    def query(self):
        """Query Game Server."""

        if self._get_a2s_info():
            self._get_a2s_players()
            if self._rules:
                self._get_a2s_rules()

        # logger.trace(f'server={self}')
        # if self._debug:
        #    pp({'server': self})

    # -------------------------------------------------------------------------------

    def _get_a2s_info(self):
        """Get A2S_INFO data."""

        addr = self.server_addr
        try:
            info = steam.game_servers.a2s_info(addr)
        except socket.timeout:
            return False
        except RuntimeError:  # as err:
            # logger.error(f'{err} a2s_info({addr})')
            return False
        except ConnectionRefusedError:  # as err:
            # logger.error(f'{err} a2s_info({addr})')
            return False

        logger.debug(f"a2s_info({addr})")
        if self._debug:
            pp({"info": info})

        #
        self.app_id = info["app_id"]
        self.server_type = info["server_type"]
        self.vac = info["vac"]
        self.visibility = info["visibility"]
        self.players = info["players"]
        self.max_players = info["max_players"]
        self.bots = info["bots"]
        self.map_name = info["map"]
        self.keywords = info["keywords"].split(",")
        self.ping = int(info["_ping"])  # from float

        # cleanup server names
        self.server_name = "".join([x for x in info["name"] if x.isprintable()]).strip()

        #
        return True

    # -------------------------------------------------------------------------------

    def _get_a2s_players(self):
        """Get A2S_PLAYER data."""

        addr = self.server_addr
        try:
            players = steam.game_servers.a2s_players(addr)
        except socket.timeout:
            return False
        except RuntimeError as err:
            logger.error(f"{err} a2s_players({addr})")
            return False

        logger.debug(f"a2s_players({addr})")
        if self._debug:
            pp({"players": players})

        self.a2s_players = sorted(players, key=lambda x: x["name"].upper())

        for player in self.a2s_players:
            if hackers := self._hackerdb.lookup_name(player["name"]):
                hacker = hackers[0]
                if not hacker.is_gamebot:
                    self.known_hackers.append(hacker)
                    logger.warning(f"hacker={hacker!r}")
                player["attributes"] = hacker.attributes_to_str()
            else:
                player["attributes"] = ""

        # legacy
        self.playernames = sorted(
            [x["name"] for x in self.a2s_players if len(x["name"]) > 0], key=lambda x: x.upper()
        )

        #
        return True

    #
    #        # Identify any known players.
    #        for name in names:
    #            if (player := self._hackerdb.lookup_name(name)):
    #                self.known_hackers.append(player)
    #                if self._debug:
    #                    print(f'found name={name} player={player!r}')
    #            elif self._debug:
    #                print(f'notfound name={name}')
    #
    #        # Check for imposters
    #        while len(names) > 0:
    #
    #            # compare the first name in the list to each of the remaining names in the list.
    #            name = names.pop(0)
    #            self.playernames.append(name)
    #
    #            keep = []
    #            for name2 in names:
    #                ratio = fuzz.ratio(name, name2)
    #                if ratio >= 80:
    #                    # looks like an imposter
    #                    logger.warning(f'ratio={ratio} name={name!r} name2={name2!r}')
    #                    self.n_imposters += 1
    #                    # remove name2 from the list.
    #                    # (nothing to do; it's effectively "removed" by not being kept
    #                    # in `keep`.)
    #                else:
    #                    if self._debug:
    #                        logger.trace(f'ratio={ratio} name={name!r} name2={name2!r}')
    #                    keep.append(name2)
    #
    #            # don't re-check/re-count identified imposters
    #            names = keep
    #
    #        #
    #        self.playernames = sorted(self.playernames, key=lambda x: x.upper())
    #
    #        #
    #        return True

    # -------------------------------------------------------------------------------

    def _get_a2s_rules(self):
        """Get A2S_RULES data."""

        addr = self.server_addr
        try:
            rules = steam.game_servers.a2s_rules(addr)
        except socket.timeout:
            return False
        except RuntimeError as err:
            logger.error(f"{err} a2s_rules({addr})")
            return False

        logger.debug(f"a2s_rules({addr})")
        if self._debug:
            pp({"rules": rules})

        #
        self.sv_tags = rules.get("sv_tags", "").split(",")

        #
        return True


# -------------------------------------------------------------------------------



"""Web app routes."""

import os
import time
from pathlib import Path

import pandas as pd
from flask import Blueprint
from flask import current_app as app
from flask import render_template, request
from loguru import logger

import qvalve.flaskapp.forms
import qvalve.gameserver
import qvalve.mainserver

# -------------------------------------------------------------------------------

bp = Blueprint("bp", __name__, template_folder="templates")

# -------------------------------------------------------------------------------


@bp.route("/", methods=("POST", "GET"))
def index():
    """Home page."""

    form = qvalve.flaskapp.forms.SearchForm()
    data = None
    datalen = 0

    if form.validate_on_submit():
        dataframe = _search(form)
        if dataframe is not None and len(dataframe) > 0:
            data = qvalve.flaskapp.forms.ServerTable(
                dataframe.to_dict("records"), table_id="servers"
            )
            datalen = len(dataframe)

    return render_template("index.html", form=form, data=data, datalen=datalen)


# -------------------------------------------------------------------------------

_MAIN_SERVER = None


def _search(form):

    try:
        global _MAIN_SERVER  # pylint: disable=global-statement
        if not _MAIN_SERVER:
            _MAIN_SERVER = qvalve.mainserver.MainServer(
                max_threads=form.max_threads.data,
                debug=form.debug.data,
            )

        servers = _MAIN_SERVER.search(
            regions=form.regions.data,
            max_servers=form.max_servers.data,
            filters=_get_query_filters(form),
        )
        if servers:
            # objs to dicts
            dataframe = pd.DataFrame([x.__dict__ for x in servers])
            dataframe = _apply_post_query_filters(form, dataframe)
            return dataframe

    except Exception as err:
        raise err
        # logger.error(err)
    return None


# -------------------------------------------------------------------------------


def _get_query_filters(form):

    filters = {"appid": form.appid.data}

    for key, value in [
        ("dedicated", form.dedicated.data),
        ("secure", form.secure.data),
        ("password", form.password.data),
        ("empty", form.empty.data),
        ("full", form.full.data),
        ("noplayers", form.noplayers.data),
    ]:
        if value == "True":
            filters[key] = 1
        elif value == "False":
            filters[key] = 0
        else:
            assert value == "None"

    if form.map_name.data:
        filters["map_name"] = form.map_name.data

    return filters


# -------------------------------------------------------------------------------


def _apply_post_query_filters(form, dataframe):

    if form.map_prefix.data:
        dataframe = dataframe[dataframe.map_name.str.startswith(form.map_prefix.data)]

    if form.min_players.data:
        dataframe = dataframe[dataframe.players > form.min_players.data]

    if form.max_bots.data:
        dataframe = dataframe[dataframe.bots <= form.max_bots.data]

    if form.max_ping.data:
        dataframe = dataframe[dataframe.ping <= form.max_ping.data]

    # sort
    # dataframe.sort_values(['mapname', 'players'], ascending=[True, False], inplace=True)
    dataframe.sort_values(
        ["map_name", "ping", "server_host", "server_port"],
        ascending=[True, True, True, True],
        inplace=True,
    )

    #    # toggle color on map change
    #    c1 = 'bg-default'
    #    c2 = 'bg-dark'
    #    c = c2
    #    lastmap = None
    #    for idx, row in dataframe.iterrows():
    #        if lastmap != row['map']:
    #            c = c1 if c == c2 else c2
    #        lastmap = row['map']
    #        dataframe.at[idx, 'tr_attrs'] = c
    #        # dataframe.at[idx, 'n_imposters'] = int(row['n_imposters'])

    return dataframe


# -------------------------------------------------------------------------------


@bp.route("/connect/<addr>", methods=("POST", "GET"))
def connect(addr):
    """Connect to game server.

    Create a tf2 `exec` script that contains a statement to connect to the
    given server. Run the script manually from tf2's console, or from a
    bound key, e.g. bind F12 "exec user/qvalve-connect.cfg"

    Args:
        addr: address of game server.

    """

    tf2_home = Path.home() / "tf2"  # create symlink if necessary
    if not tf2_home.is_dir():
        tf2_home = Path(".")
    script_name = Path("user", "qvalve-connect.cfg")
    connect_script = Path(tf2_home, "cfg", script_name)

    tag = "QVALVE "
    with open(connect_script, "w", encoding="utf-8") as file:
        print(
            "\n".join(
                [
                    f'echo {tag}{"-" * 40}',
                    f"echo {tag}{script_name}",
                    f"echo {tag}{time.asctime()}",
                    f'echo {tag}{request.args.get("server_name")}',
                    f'echo {tag}{request.args.get("map_name")}',
                    f"connect {addr}",
                ]
            ),
            file=file,
        )

    logger.success(f"Created {str(connect_script)!r} for {addr!r}")
    if app.config["args"].debug:
        os.system(f"/bin/ls -l {connect_script}; /bin/cat {connect_script}")
    return f"Connecting to {addr}"


# -------------------------------------------------------------------------------


@bp.route("/show-players/<addr>", methods=("POST", "GET"))
def show_players(addr):
    """Show players on server at `addr`."""

    logger.debug(f'server_name={request.args.get("server_name")}')
    logger.debug(f'map_name={request.args.get("map_name")}')
    logger.debug(f"addr={addr!r}")

    server = qvalve.gameserver.GameServer(addr)
    server.query()
    # logger.trace(f'server={server}')
    # for playername in server.playernames:
    # logger.trace(f'playername={playername}')
    # for hacker in server.known_hackers:
    # logger.warning(f'known_hacker={hacker}')

    jdoc = server.to_json()
    logger.debug(jdoc)
    return jdoc


# -------------------------------------------------------------------------------



"""Query Valve Main and Game Servers.

This package provides functions to 1) search Valve's MAIN server for list
of GAME servers, and 2) query GAME servers for current map, number of
players, etc.
"""



"""Flask Web Application."""

# -------------------------------------------------------------------------------

from flask import Flask
from loguru import logger

# -------------------------------------------------------------------------------


def run_web_server(args):
    """Create and run Flask Web Application."""

    app = _create_app(
        {
            "BOOTSTRAP_BOOTSWATCH_THEME": "solar",
            "SECRET_KEY": "qvalve",  # secrets.token_urlsafe(16)
            "SEND_FILE_MAX_AGE_DEFAULT": 0,  # DEV MODE ONLY
            "args": args,
        }
    )

    try:
        app.run(host="0.0.0.0")
    except Exception as err:  # pylint: disable=broad-except
        logger.critical(f"flask app failed {err}")


# -------------------------------------------------------------------------------


def _create_app(config=None):
    """Create and return Flask Web Application."""

    app = Flask(config.get("name", __name__) if config else __name__)

    if isinstance(config, dict):
        app.config.update(config)

    with app.app_context():
        # pylint: disable=import-outside-toplevel
        from flask_babel import Babel

        _ = Babel(app)
        from flask_bootstrap import Bootstrap

        _ = Bootstrap(app)

        import qvalve.flaskapp.routes

        app.register_blueprint(qvalve.flaskapp.routes.bp)
        return app


# -------------------------------------------------------------------------------
